import time
from termcolor import colored
import os
# from .jupyter_backend import create_client, run_code, start_jupyter
from .process.run import run_cell, check_for_gpu
from .process import cells

# splits a python file into an init section and a handler section
def split_file(watch):
     # split after init
    with open(watch, "r") as file:
        lines = file.readlines()
    in_init_block = False
    init_end = 0
    handler_end = 0
    for i, line in enumerate(lines):
        if "def init()" in line:
            in_init_block = True
            continue
        if 'if __name__ == "__main__":' in line:
            handler_end = i
        if in_init_block:
            # if that line is not empty, and not indented, the init block has finished
            if len(line)>0 and len(line) - len(line.lstrip()) == 0 :
                init_end = i
                in_init_block = False
                
    init_block_lines = lines[:init_end]
    handler_block_lines = lines[init_end:handler_end]

    # strip empty lines to avoid unnecessary rebuilds, then join to string
    init_block = "".join([line for line in init_block_lines if line != "\n"])
    handler_block = "".join([line for line in handler_block_lines if line != "\n"])

    return init_block, handler_block

replacements = {}
# populate_replacements gathers all patterns seen for CUDA usage in pytorch and provides a replacement value for use in str.replace()
def populate_replacements():
    replacements = {
        ".cuda()": "", 
        "device='cuda'": "device='cpu'",
        "device = 'cuda'": "device = 'cpu'",
        'device="cuda"': 'device="cpu"',
        'device = "cuda"': 'device = "cpu"'
    }

    slots = [""] # the all gpus slot
    # we only support for up to 8 GPUs, and do not support specific numbered gpus such as `cuda:1,3`
    for i in range(8):
        slots.append(f":{i}")

    # add named slots
    for slot in slots:
        replacements[f".to('cuda{slot}')"] = ""
        replacements[f'.to("cuda{slot}")'] = ""

        # devices that may be later referenced
        replacements[f"torch.device('cuda{slot}')"] = "torch.device('cpu')"
        replacements[f'torch.device("cuda{slot}")'] = "torch.device('cpu')"

    # device indices
    for i in range(8):
        replacements[f"device={i}"] = "device=-1"
        replacements[f"device = {i}"] = "device = -1"

    return replacements

# strip_cuda_calls replaces calls to cuda with calls to cpu
def strip_cuda_calls(block: str):
    # from https://pytorch.org/docs/stable/notes/cuda.html

    global replacements
    if len(replacements) == 0:
        # populate it
        replacements = populate_replacements()

    
    old_block = block
    changed = False
    for substr, new in replacements.items():
        block = block.replace(substr, new)
    if block != old_block:
        changed = True

    return block, changed

def start_all(b1, b2, first_run = False):
    if first_run:
        print(colored("------\nStarting server 🍌", 'yellow'))
    else:
        print(colored("------\nRestarting server 🍌", 'yellow'))
    # define init and handlers
    run_cell(b1)
    run_cell(b2)
    # run init
    print(colored("Running init()", 'yellow'))
    
    try:
        run_cell(cells.run_init)
    except Exception as e:
        run_cell("app.init_func()")

    # start server
    run_cell(cells.start_server)
    print(colored("Serving on http://localhost:8000\n------", 'green'))

def reload_all(b1, b2):
    # stop server then restart
    run_cell(cells.stop_server)
    start_all(b1, b2)
             
def reload_handlers(b2):
    print(colored("------\nHot reloading 🔥", 'yellow'))
    # stop server
    run_cell(cells.stop_server)
     # redefine handlers
    run_cell(b2)
    # start server
    run_cell(cells.start_server)
    print(colored("Reloaded\n------", 'green'))

# runs a hot-reload dev server
def run_dev_server(app_path, site_packages, auto_compat):
    import signal
    import sys
    # shut down thread if one is runnning
    def sigint_handler(signal, frame):
        print(colored("\nStopping server", 'yellow'))
        run_cell(cells.stop_server_quietly)
        print(colored("Bye! 👋\n------", 'green'))
        sys.exit(0)
    signal.signal(signal.SIGINT, sigint_handler)

    # DISCLAIMER: this script:
    # - assumes handler is all code after init block
    # - does not detect changes in imported files
    # - probably has a bunch of bugs

    # run session backend in virtualenv, or in global env if venv not found
    if site_packages != None:
        run_cell("import sys")
        run_cell(f"sys.path.append('{site_packages}')")
    else:
        print(colored("Warning: no virtual environment found; running in global environment", 'yellow'))

    if auto_compat:
        gpu_exists = check_for_gpu()
    compat_warning_raised = False # for a one-time warning

    # imports and classes needed for hot reload serving
    run_cell(cells.prepare_env)

    prev_b1 = 0
    prev_b2 = 0
    first_run = True
    while True:
        b1, b2 = split_file(app_path)
        
        if auto_compat and not gpu_exists:
            b1, changed_b1 = strip_cuda_calls(b1)
            b2, changed_b2 = strip_cuda_calls(b2)
            if (changed_b1 or changed_b2) and not compat_warning_raised:
                print(colored("Auto-compat warning: GPU calls have been replaced with CPU calls", 'yellow'))
                compat_warning_raised = True
                
        
        if first_run:
            start_all(b1, b2, first_run)
            first_run = False
            prev_b1 = b1
            prev_b2 = b2
            continue

        if b1 != prev_b1:
            reload_all(b1, b2)
            prev_b1 = b1
            prev_b2 = b2
            continue

        if b2 != prev_b2:
            reload_handlers(b2)
            prev_b2 = b2
        
        time.sleep(0.1)



import os
import venv
import shutil
import subprocess
import sys
import click
import git

# Uses git to clone the potassium boilerplate from /boilerplate/potassium in this git repo
def download_boilerplate(target_dir):
    # git clone to tmp dir
    temp_dir = os.path.join(target_dir, "tmp")
    git.Repo.clone_from("https://github.com/bananaml/banana-cli.git", temp_dir)
    
    # move boilerplate to current dir
    boilerplate_path = os.path.join(temp_dir, "boilerplate/potassium")
    files = os.listdir(boilerplate_path)
    for f in files:
        src_path = os.path.join(boilerplate_path, f)
        dst_path = os.path.join(target_dir, f)
        shutil.move(src_path, dst_path)
    
    # remove temp dir
    shutil.rmtree(temp_dir)

def add_git(target_dir):
    r = git.Repo.init(target_dir)
    to_commit = os.listdir(target_dir)
    to_commit.remove(".git")
    r.index.add(to_commit)
    r.index.commit("initial commit")

# Get the relative path to whatever directory the user specified
def get_target_dir(dir):
    # route to cwd if no path specified
    if len(dir) == 0:
        target_dir = "."
    else:
        target_dir = dir[0]
    # clean to relative path from here
    target_dir = os.path.relpath(target_dir)
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    return target_dir

# Get the relative path to the users app.py (or equiv. named file)
def get_app_path(entrypoint):
    # route to cwd if no path specified
    if len(entrypoint) == 0:
        app_path = "."
    else:
        app_path = entrypoint[0]
    # if dir specified, route to app.py within that dir
    dir_path = ""
    if os.path.isdir(app_path):
        dir_path = app_path
        app_path = os.path.join(dir_path, "app.py")
    # clean to relative path from here
    app_path = os.path.relpath(app_path)
    if not os.path.exists(app_path):
        raise click.UsageError("app.py not found in directory: " + os.path.abspath(dir_path) + "\n\nIf using an entrypoint other than app.py, specify it by name with `banana dev path/to/entrypoint.py`\nIf starting a new project, run `banana init`")
    return app_path

# Get the relative path to the venv site-packages
# Return None if not exists
def get_site_packages(app_path, venv_name = "venv"):
    parent_dir = os.path.split(os.path.abspath(app_path))[0]
    venv_dir = os.path.join(parent_dir, venv_name)
    if not os.path.exists(venv_dir) or not os.path.isdir(venv_dir):
        return None

    # Drill in to site packages at /{venv}/lib/python{some version}/site-packages
    lib_path = os.path.join(venv_dir, "lib")
    if not os.path.exists(lib_path):
        return None
    python_versions = os.listdir(lib_path)
    # There should only be one python version. If more, there's some trickery afoot.
    if len(python_versions) != 1:
        return None
    site_packages_dir = os.path.relpath(os.path.join(lib_path, python_versions[0], "site-packages"))
    if not os.path.exists(site_packages_dir) or not os.path.isdir(site_packages_dir):
        return None 
    
    return site_packages_dir

# create a venv
def create_venv(venv_path):
    venv.create(venv_path, with_pip=True)

# install requirements.txt into venv
def install_venv(req_path, venv_path):
    python_interpreter = os.path.join(venv_path, "bin", "python3")
    subprocess.check_call([python_interpreter, "-m", "pip", "install", "-r", req_path])


# click is our CLI library
import click
import os

# local imports
from .cmd_dev import run_dev_server
from .utils import get_target_dir, get_app_path, get_site_packages, download_boilerplate, add_git, create_venv, install_venv

@click.group()
def cli():
    pass

@click.command()
@click.argument('path', type=click.Path(exists=False, dir_okay=True, file_okay=False), nargs=-1)
@click.option('--no-venv', is_flag=True, required=False, help="Disable automatic use of a virtual environment")
@click.option('--no-install', is_flag=True, required=False, help="Disable automatic install of requirements.txt")
@click.option('--no-git', is_flag=True, required=False, help="Disable automatic creation of a git repo")
def init(path, no_venv, no_install, no_git):
    click.echo('⏰ Downloading boilerplate...')
    target_dir = get_target_dir(path)
    if target_dir == ".":
        # verify they're not installing into an existing dir
        if len(os.listdir(target_dir)) > 0:
            click.confirm("This is not an empty directory, and the install may overwrite existing files.\nDo you want to proceed?", abort=True)
    download_boilerplate(target_dir)
    if not no_git:
        add_git(target_dir)
    if not no_venv:
        click.echo('🌎 Creating virtual environment...')
        venv_path = os.path.join(target_dir, "venv")
        create_venv(venv_path)

        if not no_install:
            click.echo('📦 Downloading packages...')
            req_path = os.path.join(target_dir, "requirements.txt")
            install_venv(req_path, venv_path)

    click.echo('\n🍌 Project ready to go (hurrah!)')
    click.echo('\n🔥 To run a dev server with hot-reload, run:')
    if target_dir != ".":
        click.echo(f'cd {target_dir}')
    if no_install:
        click.echo(f'banana install')
    click.echo('banana dev')

@click.command()
@click.option('--venv', default="venv", required=False, type=str, help="The path of the virtual environment to install into. Defaults to venv.")
def install(venv):
    if not os.path.exists(venv):
        click.echo('🌎 Creating virtual environment...')
        create_venv(venv)
    click.echo('📦 Downloading packages...')
    install_venv("requirements.txt", venv)

@click.command()
@click.option('--venv', default="venv", required=False, type=str, help="The path of the virtual environment to run in. Defaults to venv.")
@click.option('--auto-compat', default=False, required=False, type=bool, help="Prevents cuda use if there is no GPU visible. Defaults to True.")
@click.argument('entrypoint', type=click.Path(exists=True), nargs=-1)
def dev(venv, auto_compat, entrypoint):
    app_path = get_app_path(entrypoint)
    site_packages = get_site_packages(app_path, venv_name = venv)
    run_dev_server(app_path, site_packages, auto_compat)

cli.add_command(init)
cli.add_command(install)
cli.add_command(dev)

if __name__ == "__main__":
    cli()



def run_cell(input_file):
    exec(input_file, globals())

def check_for_gpu():
    try:
        exec('''
import torch
torch.tensor([1.,]).cuda()
        ''', globals())
        return True
    except:
        return False



prepare_env = '''
import threading, os
from werkzeug.serving import make_server

class ServerThread(threading.Thread):
    def __init__(self, app):
        threading.Thread.__init__(self)
        self.server = make_server('127.0.0.1', 8000, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        self.server.serve_forever()

    def stop(self):
        self.server.shutdown()
'''

run_init = "app._init_func()"

start_server = '''
flask_app = app._create_flask_app()
server = ServerThread(flask_app)
server.start()
'''

stop_server = "server.stop()"

stop_server_quietly = '''
try:
    server.stop()
except:
    pass
'''
